#!/usr/bin/python
from __future__ import print_function
import time
import calendar
import subprocess
import signal
import os
import sys
import json
import logging
from pidHandler import getProcessStatus, writePidFile, deletePidFile
import boDb
from boConfig import SB_INTERFACE, SB_IP, LOGLEVEL
import lzf

# other stuff
NTP_PORT = 123
DNS_PORT = 53
CHILD_PIDNAME = "boUdpParser-tcpdump" # pid of the child process tcpdump
MY_PIDNAME = "boUdpParser"
LZF_MAGIC = ord('Z')
frame = None # currently collected IP frame
fragment = {} # IP fragment, indexed by destination IP address
logger = None

class Namespace: pass # workaround to create a shared namespace for functions and functions in functions

def configLogger():
    global logger
    format = "%(asctime)-15s %(levelname)s: %(message)s"
    filename = "/var/log/boUdpParser.log"
    logging.basicConfig(filename = filename, format = format)
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)
    logger.info("log started; pid = {}".format(os.getpid()))
    logger.setLevel(LOGLEVEL)

def signal_handler(signal, frame):
    boDb.disconnectMysql()
    logger.error('program terminated!')
    sys.exit(0)

def startFrame(line):
    global frame
    logger.info(line)
    if ("tcpdump" in line) or ("listening" in line):
        frame = None
    else:
        frame = {}
        frame["data"] = []

def continueFrame(line):
    global frame
    if ("0x" == line[0:2]) and (":  " == line[6:9]):
        offset = int(line[2:6], 16)
        words = line[9:]
        words = words.split()
        bytes = []
        for i in words:
            if (2 == len(i)):
                bytes.append(i)
            elif (4 == len(i)):
                bytes.append(i[0:2])
                bytes.append(i[2:])
            else:
                logger.error("len(i) is unexpected '{}'".format(i))
                frame = None
        bytes = [int(i, 16) for i in bytes]
        if (offset == len(frame["data"])):
            frame["data"] += bytes
        else:
            logger.error("wrong offset, expected = 0x{:04X}, found = 0x{:04X}".format(len(frame["data"]), offset))
            frame = None
    else:
        logger.error("line '{}' not supported".format(line))
        frame = None

def checkUdpFrame(udp):
    isOk = False # bad frame
    srcPort = udp[0] * 0x100 + udp[1]
    dstPort = udp[2] * 0x100 + udp[3]
    length = udp[4] * 0x100 + udp[5]
    checksum  = udp[6] * 0x100 + udp[7]
    if (length == len(udp)):
        # TODO: verify UDP checksum
        if (NTP_PORT == srcPort) and (NTP_PORT == dstPort):
            # the frame may be valid, but we drop it
            logger.info("NTP frame dropped")
            isOk = None # frame was ok but is not to be regarded
        elif (DNS_PORT == dstPort):
            # the frame may be valid, but we drop it
            logger.info("DNS frame dropped")
            isOk = None # frame was ok but is not to be regarded
        else:
            isOk = True # good frame
    else:
        logger.error("length = {} different to len(udp) = {}".format(length, len(udp)))
    return isOk

def dumpUdpFrame(udp):
    srcPort = udp[0] * 0x100 + udp[1] # don't care
    dstPort = udp[2] * 0x100 + udp[3] # don't care
    length = udp[4] * 0x100 + udp[5]
    checksum  = udp[6] * 0x100 + udp[7]
    logger.debug("srcPort = {}, dstPort = {}, length = 0x{:04X}, checksum = 0x{:04X}".format(srcPort, dstPort, length, checksum))
    data = udp[8:]
    text = ""
    hex = ""
    chars = ""
    for i in range(len(data)):
        if (0 != i) and (0 == i % 16):
            text += hex + "    " + chars + "\n"
            hex = ""
            chars = ""
        hex += "{:02X} ".format(data[i])
        if (data[i] >= 32) and (data[i] < 127):
            chars += chr(data[i])
        else:
            chars += "."
    if (len(hex)):
        text += hex + " " * (52 - len(hex)) + chars
    logger.debug(text)

def decompress(zip):
    decompressed = ""
    if (LZF_MAGIC == zip[0]):
        zip_str = ''.join(chr(i) for i in zip[1:]) # skip the 'Z' at position 0
        try:
            decompressed = lzf.decompress(zip_str, 8192)
            logger.info("lzf.decompressed(udp 'Z' frame)")
        except:
            try:
                decompressed = lzf.decompress(zip_str + '\n', 8192)
                logger.info("lzf.decompressed(udp 'Z' frame appended by '\\n')")
            except:
                logger.error("lzf.decompressed(udp 'Z' frame failed')")
    else:
        logger.error("unexpected LZF_MAGIC 0x{:02X}".format(zip[0]))
    decompressed = [ord(i) for i in decompressed]
    return decompressed

def parseUdpFromSystemBlue(udp):
    ns = Namespace()
    signal = {'channel': [{}, {}, {}, {}]} # a dictionary with an empty array of dictionaries for four channels; System Blue has 4 channels: 0 = 1A, 1 = 1B, 2 = 1C, 3 = 2A
    ns.dataChannel = None # channel for the collection of the binary data
    ns.dataLength = 0 # length of the binary data to be collected

    def parseTextLine(line):
        """returns a dictionary with a key and it's value, where the vlaue may be a dictionary itself"""
        sl = line.split()
        key = sl[0]
        if ("station" == key) and (2 == len(sl)):
            stationNumber = sl[1]
            signal[key] = stationNumber
        elif ("processor" == key) and (2 == len(sl)):
            deviceId = sl[1]
            signal[key] = deviceId
        elif ("time" == key) and (3 == len(sl)):
            datetime = sl[1] + " " + sl[2].split(".")[0]
            nanosec = sl[2].split(".")[1]
            t = time.strptime(datetime, "%Y-%m-%d %H:%M:%S")
            sec = int(calendar.timegm(t))
            signal[key] = sl[1] + ' ' + sl[2]
            signal['time_ns'] = sec * 1000000000 + int(nanosec)
        elif ("position" == key) and (4 == len(sl)):
            lat = float(sl[1])
            lon = float(sl[2])
            height = int(sl[3])
            signal['lat'] = lat
            signal['lon'] = lon
            signal['height'] = height
        elif ("amp" == key) and (6 == len(sl)):
            channel = int(sl[1])
            preAmp = sl[2]
            gain = sl[3]
            val1 = int(sl[4]) # meaning unknown
            val2 = int(sl[4]) # meaning unknown
            signal['channel'][channel][key] = {'preAmp': preAmp, 'gain': gain, 'val1': val1, 'val2': val2}
        elif ("conversion_time" == key) and (3 == len(sl)):
            channel = int(sl[1])
            val = int(sl[2]) # meaning unknown
            signal['channel'][channel][key] = val
        elif ("conversion_gap" == key) and (3 == len(sl)):
            channel = int(sl[1])
            val = int(sl[2]) # meaning unknown
            signal['channel'][channel][key] = val
        elif ("values" == key) and (3 == len(sl)):
            channel = int(sl[1])
            val = int(sl[2]) # meaning unknown
            signal['channel'][channel][key] = val
        elif ("start" == key) and (3 == len(sl)):
            channel = int(sl[1])
            val = int(sl[2]) # meaning unknown
            signal['channel'][channel][key] = val
        elif ("shift" == key) and (3 == len(sl)):
            channel = int(sl[1])
            val = int(sl[2]) # meaning unknown
            signal['channel'][channel][key] = val
        elif ("coding" == key) and (3 == len(sl)):
            channel = int(sl[1])
            val = int(sl[2]) # meaning unknown
            signal['channel'][channel][key] = val
        elif ("data" == key) and (4 == len(sl)) and ('$' == sl[3]):
            ns.dataChannel = int(sl[1]) # collect the following binary data for this channel
            ns.dataLength = int(sl[2]) # expected length of the binary data
            signal['channel'][ns.dataChannel]['data'] = []
        else:
            logger.error("unhandled text line '{}'".format(line))
            logger.error(frame)

    dumpUdpFrame(udp)
    checkResult = checkUdpFrame(udp)
    if (True == checkResult):
        if ('Z' == chr(udp[8])): # first magic data byte which announces the 'zipped' format; the actual compression mechanism is unknown
            #logger.warning("TODO: interpret 'Z' frames")
            #return None
            udp = udp[0:8] + decompress(udp[8:])
            if (8 < len(udp)):
                dumpUdpFrame(udp)
            else:
                logger.error("can't handle compressed udp frame")
                return

        text = "" # temporary variable used to collect ascii data

        # state variables for the parsing
        stateCollectText = 1 # state number
        stateCollectData = 2 # state number
        stateEndOfData = 3 # state number
        state = stateCollectText # state variable

        for i in udp[8:]: # UDP data starts at offset 8
            if (stateCollectText == state): # first phase
                text += chr(i)
                if '$' == chr(i):
                    lines = text.split('\n')
                    ns.dataChannel = None # has to be set by parseTextLine() if a valid 'data' line has been detected
                    ns.dataLength = 0 # has to be set by parseTextLine() if a valid 'data' line has been detected
                    for line in lines:
                        parseTextLine(line)
                    if (None != ns.dataChannel) and (0 != ns.dataLength):
                        text = "" # clear text buffer in order to start with an empty buffer the next time stateCollectText is entered
                        state = stateCollectData # continue with second phase
                    else:
                        logger.error("unexpected last line for a data block starting with '$' '{}'".format(lines[-1]))
            elif (stateCollectData == state): # second phase
                signal['channel'][ns.dataChannel]['data'].append(i)
                if (ns.dataLength == len(signal['channel'][ns.dataChannel]['data'])):
                    state = stateEndOfData # continue with third phase
            elif (stateEndOfData == state): # third phase
                if ('\n' == chr(i)):
                    state = stateCollectText # continue with first phase
                    if (logging.DEBUG == logger.level):
                        logger.debug("".join("%02X " %x for x in signal['channel'][ns.dataChannel]['data']))
            else:
                logger.error("illegal state while parsing UDP frame from System Blue")
                sys.exit(1)
        return signal
    elif (None == checkResult):
        # dropped frame i.e. NTP
        return None
    else:
        logger.error("invalid UDP frames")
        return None

def isChecksumOk(data):
    checkSum = 0x00000000 # startValue
    #sum up 16 bit values
    while (2 <= len(data)):
        checkSum += data[0] * 0x100 + data[1]
        data = data[2:]
    # reminder is just 8 bit, still treat it as 16 bit vaue by appending it with 0x00
    if (1 == len(data)):
        checkSum += data[0] * 0x100 + 0x00
        data = data[1:]
    # treat carry flags in the overflow
    while (0xFFFF < checkSum):
        checkSum = (checkSum & 0xFFFF) + (checkSum / 0x10000)
    return (0xFFFF == checkSum) # the valid result is 0x0000

# this is a pretty shortcut implementtation
# it handles just two fragments which have to appear in order
def handleIpFragment(moreFragments, fragmentOffset, dstAddress):
    global fragment
    global frame
    data = []
    if (0 == fragmentOffset):
        if (1 == moreFragments):
            # first fragment
            fragment[dstAddress] = frame
        else:
            logger.error("unexpected: dstAddress = {}, fragmentOffset = {}, moreFragments = {}".format(dstAddress, fragmentOffset, moreFragments))
            logger.error("frame = {}".format(frame))
            fragment[dstAddress] = None
    else: # (0 < fragmentOffset)
        # second or following fragment
        if (fragment[dstAddress]):
            # verify version and IHL match,
            # identification matches
            # protocol matches
            # source address matches
            # destination address matches
            if (fragment[dstAddress]['data'][0] == frame['data'][0]) and \
               (fragment[dstAddress]['data'][4:6] == frame['data'][4:6]) and \
               (fragment[dstAddress]['data'][9] == frame['data'][9]) and \
               (fragment[dstAddress]['data'][12:16] == frame['data'][12:16]) and \
               (fragment[dstAddress]['data'][16:20] == frame['data'][16:20]):
                ihl = fragment[dstAddress]['data'][0] & 0x0F
                fragmentHeaderLen = 4 * ihl
                fragmentPayloadLen = len(fragment[dstAddress]['data']) - fragmentHeaderLen # has to be a multiple of 8
                if (0 == (fragmentPayloadLen % 8)):
                    if (fragmentPayloadLen == (fragmentOffset * 8)):
                        fragment[dstAddress]['data'] += frame['data'][4*ihl:]
                        if (0 == moreFragments):
                            # the last fragment
                            data = fragment[dstAddress]['data']
                            fragment[dstAddress] = None
                    else:
                        logger.error("fragment offset does not match previous fragment payload length")
                        fragment[dstAddress] = None
                else:
                    logger.error("previous fragment payload length = {} is not a multiple of 8".format(fragmentPayloadLen))
                    fragment[dstAddress] = None
            else:
                logger.error("fragment headers do not match")
                fragment[dstAddress] = None
        else:
            logger.error("more fragments without first fragment")
            fragment[dstAddress] = None
    frame = None
    return data

def parseIpFrame():
    global frame
    data = frame["data"]
    # some plausibility checks for IPv4
    version = ((data[0] >> 4) & 0x0F)
    if (4 == version):
        ihl = data[0] & 0x0F
        if (5 <= ihl):
            ipHeader = data[0:4*ihl]
            if isChecksumOk(ipHeader):
                #headerChecksum = ipHeader[10] * 0x100 + ipHeader[11] # value as such not important, it is already checked with isChecksumOk()
                srcAddress = "{}.{}.{}.{}".format(ipHeader[12], ipHeader[13], ipHeader[14], ipHeader[15])
                dstAddress = "{}.{}.{}.{}".format(ipHeader[16], ipHeader[17], ipHeader[18], ipHeader[19]) # don't care
                if (srcAddress == SB_IP):
                    #tos = ipHeader[1] # don't care
                    totalLength = ipHeader[2] * 0x100 + ipHeader[3]
                    if (totalLength < len(data)) and (46 == len(data)):
                        # Short IP frames with less than 46 byte, have been observed to be padded.
                        # The reason is the minimum size of a Layer 2 Ethernet frame of 64 byte.
                        # The 64 byte are calculated from:
                        #   6 octets MAC destination
                        #   6 octets MAC source
                        #   2 octests Ethertype (Ethernet II) or length (IEEE 802.3)
                        #  46 octets Payload (here this is the IP frame)
                        #   4 octets Frame check sequence (32 bit CRC)
                        paddingLength = len(data) - totalLength
                        frame["data"] = frame["data"][0:-paddingLength]
                        data = data[0:-paddingLength]
                        logger.info("{} bytes of padding removed".format(paddingLength))
                    if (totalLength == len(data)):
                        identification = ipHeader[4] * 0x100 + ipHeader[5] # don't care
                        flags = (ipHeader[6] >> 5) & 0x07
                        reserved = (flags >> 2) & 0x01 # has to be 0
                        dontFragment = (flags >> 1) & 0x01 # doesn't matter for this sniffer
                        moreFragments = (flags >> 0) & 0x01
                        if (0 == reserved):
                            fragmentOffset = (ipHeader[6] & 0x1F) * 0x100 + ipHeader[7]
                            if (0 != moreFragments) or (0 != fragmentOffset):
                                data = handleIpFragment(moreFragments, fragmentOffset, dstAddress)
                            if (data):
                                ttl = ipHeader[8] # don't care
                                protocol = ipHeader[9]
                                if (17 == protocol):
                                    #TODO: if ihl > 5 check for options
                                    # end of ipHeader evaluation
                                    # start of UDP
                                    udp = data[4*ihl:]
                                    return parseUdpFromSystemBlue(udp)
                                else:
                                    logger.error("protocol = {} not supported, UDP = 17 expected".format(protocol))
                        else:
                            logger.error("reserved flag is set flags = {}".format(flags))
                    else:
                        logger.error("len(data) = {}, totalLength = {}".format(len(data), totalLength))
                else:
                    logger.error("src ip = {} is not expected; expected is src ip = {}".format(srcAddress, SB_IP))
            else:
                logger.error("ip header '{}' checkSum is wrong".format(''.join("%02X " %x for x in ipHeader)))
        else:
            logger.error("ihl = 0x{:X} too short)".format(ihl))
    else:
        logger.error("IPv{} not supported".format(version))
    return None

def captureInput(interface, ip):
    global frame
    cmd = "tcpdump -i {} -x -n udp and src host {}".format(interface, ip)
    logger.info(cmd)
    sp = subprocess.Popen(cmd.split(), stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

    # Write PID file
    writePidFile(CHILD_PIDNAME, sp.pid)

    while True:
        # Check if the subprocess is still running
        status, pid = getProcessStatus(CHILD_PIDNAME)
        if (None == status) or (None == pid) or (sp.pid != pid) or ('zombie' == status):
            logger.error("'{}' status = {}, pid = {}".format(CHILD_PIDNAME, status, pid))
            sys.exit(1)

        line = sp.stdout.readline()
        line = line.strip()
        if line:
            logger.debug(line)
            if ("0x" != line[0:2]):
                if frame:
                    signal = parseIpFrame()
                    if (None != signal):
                        signalHasData = False
                        for channel in range(4):
                           if ('data' in signal['channel'][channel]):
                               signal['channel'][channel]['data'] = "0x{:02X} ...".format(signal['channel'][channel]['data'][0])
                               signalHasData = True
                        if signalHasData:
                            logger.debug(json.dumps(signal, indent=4))
                            if not boDb.insertSignal(signal):
                                logger.error("boDb.insertSignal() failed with signal = {}".format(json.dumps(signal, indent=4)))
                        else:
                            logger.error("frame = {}".format(frame))
                            logger.error("signal = {}".format(json.dumps(signal, indent=4)))
                    frame = None
                startFrame(line)
            else:
                continueFrame(line)
        else:
            time.sleep(1)

def startup():
    global logger

    configLogger()

    cmd = ["which", "tcpdump"]
    sp = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    out, err = sp.communicate()
    if ("tcpdump" in out) and ("" == err):
        signal.signal(signal.SIGINT, signal_handler)
        boDb.connectMysql()
        boDb.connectDatabase()
        captureInput(SB_INTERFACE, SB_IP)
    else:
        if err:
            logger.error("'which tcpdump': '{}'".format(err))
        logger.error("tcpdump seems not to be installed, please execute 'sudo apt-get install tcpdump'")

def supervise():
    # Check if another instance is still running
    status, pid = getProcessStatus(MY_PIDNAME)
    if (None == status) or (None == pid) or ('zombie' == status):
        # Write PID file
        pid = os.getpid()
        writePidFile(MY_PIDNAME, os.getpid())
        # not yet running, continue
        startup()
    else:
        # another instance is running, terminate
        sys.exit(0)

def main():
    supervise()

if __name__ == "__main__":
    main()
